/**
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [https://neo4j.com]
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* eslint-disable @typescript-eslint/no-unused-vars */

import {
  Integer,
  Record,
  ResultSummary,
  QueryResult,
  Result,
  Transaction,
  Session,
  TransactionConfig,
  Node,
  Relationship
} from 'neo4j-driver-core'

const dummy: any = null
const intValue: Integer = Integer.fromInt(42)

const session: Session = dummy

const txConfig1: TransactionConfig = {}
const txConfig2: TransactionConfig = { timeout: 5000 }
const txConfig3: TransactionConfig = { timeout: intValue }
const txConfig4: TransactionConfig = { metadata: {} }
const txConfig5: TransactionConfig = {
  metadata: {
    key1: 'value1',
    key2: 5,
    key3: { a: 'a', b: 'b' },
    key4: [1, 2, 3]
  }
}
const txConfig6: TransactionConfig = {
  timeout: 2000,
  metadata: { key1: 'value1', key2: 2 }
}
const txConfig7: TransactionConfig = {
  timeout: intValue,
  metadata: { key1: 'value1', key2: 2 }
}

const tx1: Transaction = session.beginTransaction()
const bookmarks: string[] = session.lastBookmarks()

const promise1: Promise<number> = session.readTransaction((tx: Transaction) => {
  return 10
})

const promise2: Promise<string> = session.readTransaction(async (tx: Transaction) => {
  return await Promise.resolve('42')
})

const promise3: Promise<number> = session.writeTransaction(
  (tx: Transaction) => {
    return 10
  }
)

const promise4: Promise<string> = session.writeTransaction(
  async (tx: Transaction) => {
    return await Promise.resolve('42')
  }
)

const close1: Promise<void> = session.close()
const close2: Promise<void> = session.close().then(() => {
  console.log('Session closed')
})

const result1: Result = session.run('RETURN 1')
result1
  .then((res: QueryResult) => {
    const records: Record[] = res.records
    const summary: ResultSummary = res.summary
    console.log(records)
    console.log(summary)
  })
  .catch((error: Error) => {
    console.log(error)
  })

const result2: Result = session.run('RETURN 2')
result2.subscribe({})
result2.subscribe({
  onNext: (record: Record) => console.log(record)
})
result2.subscribe({
  onNext: (record: Record) => console.log(record),
  onError: (error: Error) => console.log(error)
})
result2.subscribe({
  onNext: (record: Record) => console.log(record),
  onError: (error: Error) => console.log(error),
  onCompleted: (summary: ResultSummary) => console.log(summary)
})

const result3: Result = session.run('RETURN $value', { value: '42' })
result3
  .then((res: QueryResult) => {
    const records: Record[] = res.records
    const summary: ResultSummary = res.summary
    console.log(records)
    console.log(summary)
  })
  .catch((error: Error) => {
    console.log(error)
  })

const result4: Result = session.run('RETURN $value', { value: '42' })
result4.subscribe({})
result4.subscribe({
  onNext: (record: Record) => console.log(record)
})
result4.subscribe({
  onNext: (record: Record) => console.log(record),
  onError: (error: Error) => console.log(error)
})
result4.subscribe({
  onNext: (record: Record) => console.log(record),
  onError: (error: Error) => console.log(error),
  onCompleted: (summary: ResultSummary) => console.log(summary)
})

const result5: Result = session.run('RETURN $value', { value: '42' }, txConfig1)
result5
  .then((res: QueryResult) => {
    const records: Record[] = res.records
    const summary: ResultSummary = res.summary
    console.log(records)
    console.log(summary)
  })
  .catch((error: Error) => {
    console.log(error)
  })

const result6: Result = session.run('RETURN $value', { value: '42' }, txConfig2)
result6.subscribe({})
result6.subscribe({
  onNext: (record: Record) => console.log(record)
})
result6.subscribe({
  onNext: (record: Record) => console.log(record),
  onError: (error: Error) => console.log(error)
})
result6.subscribe({
  onNext: (record: Record) => console.log(record),
  onError: (error: Error) => console.log(error),
  onCompleted: (summary: ResultSummary) => console.log(summary)
})

const tx2: Transaction = session.beginTransaction(txConfig2)
const promise5: Promise<string> = session.readTransaction(
  (tx: Transaction) => '',
  txConfig3
)
const promise6: Promise<number> = session.writeTransaction(
  (tx: Transaction) => 42,
  txConfig4
)

const lastBookmarks: string[] = session.lastBookmarks()

interface Person {
  age: Integer
  name: string
}

interface Friendship {
  since: Integer
}

interface PersonAndFriendship {
  p: Node<number, Person>
  f: Relationship<number, Friendship>
}

const personSessionRun = session.run<Person>('MATCH (p:Person) RETURN p.name, p.age')
personSessionRun.then(({ records }) => {
  for (const person of records) {
    const age: Integer = person.get('age')
    const name: string = person.get('name')

    // @ts-expect-error
    const nameInt: Integer = person.get('name')
  }
}).catch(err => console.error(err))

const personAndFriend = session.run<PersonAndFriendship>('MATCH (p:Person)-[f:Friendship]-() RETURN p, f')
personAndFriend.then(({ records }) => {
  for (const r of records) {
    const person = r.get('p')

    const age: Integer = person.properties.age
    const name: string = person.properties.name

    // @ts-expect-error
    const nameInt: Integer = person.properties.name

    // @ts-expect-error
    const err: string = person.properties.err

    const friendship = r.get('f')

    const since: Integer = friendship.properties.since

    // @ts-expect-error
    const sinceString: string = friendship.properties.since

    // @ts-expect-error
    const err2: string = friendship.properties.err
  }
}).catch(error => console.error(error))

const personExecuteRead = session.executeRead(tx => tx.run<Person>('MATCH (p:Person) RETURN p.name, p.age'))

personExecuteRead.then(({ records }) => {
  for (const person of records) {
    let age: Integer = person.get('age')
    let name: string = person.get('name')

    // @ts-expect-error
    const nameInt: Integer = person.get('name')

    const p = person.toObject()

    age = p.age
    name = p.name
  }
}).catch(err => console.error(err))

const personExecuteWrite = session.executeWrite(tx => tx.run<Person>('MATCH (p:Person) RETURN p.name, p.age'))

personExecuteWrite.then(({ records }) => {
  for (const person of records) {
    const age: Integer = person.get('age')
    const name: string = person.get('name')

    // @ts-expect-error
    const nameInt: Integer = person.get('name')
  }
}).catch(err => console.error(err))
